class ModelNotFoundException(Exception):
    """Raise when a model file is not found."""

    pass


class ModelFileTypeException(Exception):
    """Raise when a model file is not json."""

    pass


class MissingModelException(Exception):
    """Raise when the model is missing"""

    pass


class MissingConfigsException(Exception):
    """Raise this when the config file missing"""

    pass


class TemplateNotFoundException(Exception):
    """Raise when a template file is not found."""

    pass


class NotFolderException(Exception):
    """Raise when a Path is not a folder."""

    pass


class BadArgumentsException(Exception):
    """Raise the passed arguments are incorect"""

    pass



import json
from pathlib import Path

from koodu.exceptions import (
    MissingConfigsException,
    ModelFileTypeException,
    ModelNotFoundException,
    NotFolderException,
)
from koodu.generator.generator import Generator


def generate(args):
    if not Path(args.output).is_dir():
        raise NotFolderException("The Output should be a Folder!")

    # check if the input template path is a whole path
    if "/" in args.templates or "\\" in args.templates:
        template_path = Path(args.templates)
    else:
        template_path = (
            Path(__file__).parent.parent / Path("templates") / Path(args.templates)
        ).resolve()

    if not Path(template_path / Path("config.yaml")).is_file():
        raise MissingConfigsException("NOT TEMPLATE CONFIG FILE")

    if not template_path.is_dir():
        raise NotFolderException(f"{args.templates} is not an Existing directory")

    if not args.model.endswith(".json"):
        raise ModelFileTypeException("The model should be a json file!")

    if not Path(args.model).is_file():
        raise ModelNotFoundException(f"{args.model} is not an Existing file")

    with open(args.model, "r") as f:
        model = json.load(f)

    if model is None:
        raise ModelFileTypeException(f"{args.model} is not a valid file")

    generator = Generator(
        model=model, template_folder=template_path, output=Path(args.output)
    )

    for file in generator.render():
        file.write()

    print("Done", "\U00002705")



from pathlib import Path
from typing import Dict

from koodu.scripts.utils import (
    check_all_template,
    convert_symbols,
    get_files_from_folder,
)


def check_templates(args) -> None:
    if not Path(args.templates).is_dir():
        raise Exception(f"{args.templates} is not an Existing directory")

    templates = get_files_from_folder()
    valids = check_all_template(templates)
    symbols = convert_symbols(valids)
    output = []
    for path, status in zip(templates, symbols):
        output.append(f"{path} {status}")
    return output


def list_command(args) -> Dict[str, str]:
    output = []
    if args.option == "models" or args.option == "m":
        model_path = Path(Path("./koodu/models").resolve())

        models = [
            elem
            for elem in model_path.glob("*")
            if str(elem).endswith(".json") or str(elem).endswith(".yaml")
        ]
        for model in models:
            name = model.stem
            output.append(f"{name}: {model}")

    elif args.option == "templates" or args.option == "t":
        template_path = Path("./koodu/templates").resolve()

        templates = [elem for elem in template_path.glob("*") if elem.is_dir()]
        for template in templates:
            name = template.stem
            output.append(f"{name}: {template}")

    return output



from pathlib import Path


class File(object):
    """The File Classe is the representation of a file."""

    def __init__(
        self,
        name: str,
        root: Path = Path("./"),
        content: str = "",
        is_binary: bool = False,
    ) -> None:
        self.name = name
        self.root = root
        self.content = content
        self.is_binary = is_binary

    def write(self):
        if not self.is_binary:
            if not self.root.is_dir():  # if the parent is not directory
                self.root.mkdir(exist_ok=True, parents=True)  # create it
            with open(self.root / self.name, "w") as fp:
                fp.write(self.content)
        else:
            pass

    def to_dict(self):
        return {
            "name": self.name,
            "root": self.root,
            "content": self.content,
            "is_binary": self.is_binary,
        }



from pathlib import Path
from typing import Dict

import yaml

from koodu.generator.models import ConfigSchema


def get_all_files(folder: Path) -> Dict[str, Path]:
    """Iterate over the folder and return all file as Path.

    Args:
        folder (Path): the path to the folder to iterate on.

    Returns:
        (dict): a dict where the key are template's name and the value are template's code
    """
    result = {}
    for fil in folder.glob():
        if fil.is_file():
            result[fil.stem] = fil

    return result


def load_template_config(folder: Path) -> ConfigSchema:
    """Read the template config file.

    Args:
        folder (Path): the path to the folder to iterate on.

    Returns:
        (dict): a dict where the key are template's name and the value are template's code
    """
    configs = None
    config_path = folder / Path("config.yaml")
    with open(config_path, "r") as fp:
        configs = yaml.load(fp, Loader=yaml.FullLoader)

    return ConfigSchema(**configs)



from typing import List, Optional, Union

from pydantic import BaseModel, ValidationError, validator

from koodu.generator.enums import ModelTypeEmum


class AttributSchema(BaseModel):
    name: str
    type: ModelTypeEmum
    size: Optional[int] = None
    primary_key: bool = False
    index_key: bool = False
    unique_key: bool = False
    required: bool = False
    model: str = None

    @validator("model", pre=True, always=True)
    def check_required_model(cls, value, values):
        attribute_type = values.get("type")
        if attribute_type == ModelTypeEmum.ref and not value:
            raise ValidationError("Model is required when the type is reference")
        return value

    @validator("size", pre=True, always=True)
    def check_required_size(cls, value, values):
        attribute_type = values.get("type")
        if attribute_type == ModelTypeEmum.string and not value:
            raise ValidationError("Size is required when the type is String")
        return value


class ModelSchema(BaseModel):
    name: str
    attributs: List[AttributSchema]

    @validator("attributs", pre=True, always=True)
    def check_reference_size(cls, values):
        models = list(map(lambda x: x["name"], values))
        for value in values:
            if value["type"] == ModelTypeEmum.ref and value["model"] not in models:
                raise ValidationError("Ref model does not exist")
        return values


class TemplateSchema(BaseModel):
    template_code: str
    name: str
    path: str
    file_name: str
    type: Union[str, None] = None
    file_path: str


class GeneratorModelSchema(BaseModel):
    name: str
    models: List[ModelSchema]


class TemplateConfigSchema(BaseModel):
    template_path: str
    file_path: Union[str, None] = None
    path: Union[str, None] = None
    name: str
    file_name: Union[str, None] = None
    type: Union[str, None] = None
    is_base: Union[bool, None] = False
    is_macro: Union[bool, None] = False


class ConfigSchema(BaseModel):
    name: Union[str, None] = None
    templates: List[Union[TemplateConfigSchema, None]] = None



"""The code generator module"""

import logging
from pathlib import Path
from typing import Dict, List

from jinja2 import (
    DictLoader as JDictLoader,
)
from jinja2 import (
    Environment as JEnvironment,
)
from jinja2 import (
    Template as JTemplate,
)
from jinja2.exceptions import TemplateNotFound

from koodu.exceptions import MissingModelException
from koodu.generator.file import File
from koodu.generator.models import GeneratorModelSchema, TemplateSchema
from koodu.generator.utils import (
    ConfigSchema,
    load_template_config,
)


class SingleModelView:
    """Convenience object to generate an object out a of a dict"""

    def __init__(self, model):
        self.elem = model

    def __str__(self):
        return str(tuple(self))

    def __iter__(self):
        for item in self.elem:
            yield item


class Generator:
    """The Generator is the Core of the engine, it takes as parameters a model, the
    path to the template folder, and the path of the folder where the generated
    code will be saved and it contains different methods that allow us to generate
    our code from the model and templates.
    """

    def __init__(
        self,
        template_folder: Path,
        model: GeneratorModelSchema = None,
        output: Path = Path("./"),
    ) -> None:
        self.valid = False
        self.last_error = ""
        self.template_folder = template_folder
        self.model = model
        self.output = output
        self.configs: ConfigSchema = load_template_config(self.template_folder)

        if self.configs is None:
            raise Exception("The config file is not good formated.")

        if self.model is None:
            raise MissingModelException("NO_MODEL", "Missing Model.")
        elif self.configs.templates is None:
            raise Exception("NO_TEMPLATES", "Missing Template")
        elif len(self.configs.templates) < 1:
            raise Exception("NO_ATTACHED_TEMPLATES", "Missing Attached Template.")
        else:
            templates = self.get_templates()
            self.jinja_env = JEnvironment(loader=JDictLoader(templates))
            self.valid = True

    def __str__(self):
        return f"""Configs: {self.configs}
        Templates: {self._list_templates()}
        """

    def get_templates(self) -> Dict[str, Dict[str, str]]:
        """Load the templates Informations."""
        result = {}
        for template in self.configs.templates:
            with open(self.template_folder / Path(template.template_path), "r") as fp:
                template_code = fp.read()

            result[template.name] = template_code

        return result

    def get_template_by_name(self, template_name: str) -> TemplateSchema:
        """Find a Template from the list of template with it name."""
        found_template = None

        for template in self.configs.templates:
            if template.name == template_name:
                with open(
                    self.template_folder / Path(template.template_path), "r"
                ) as fp:
                    template_code = fp.read()

                found_template = TemplateSchema(
                    **{
                        "template_code": template_code,
                        "name": template.name,
                        "path": template.path,
                        "file_name": template.file_name,
                        "type": template.type,
                        "file_path": template.file_path,
                    }
                )
                break
        return found_template

    def print_templates(self) -> Dict[str, Dict[str, str]]:
        """Get the liste of Template"""
        return self.get_templates()

    def render_file_name(self, template: TemplateSchema, model) -> str:
        """Render file name

        Args:
            template (dict): The Template to generate.
            model (str): The model from which we want to generate.

        Returns:
            (str): the file name
        """
        if len(template.file_name) > 0:
            args = {"model": model, "full_model": self.model}
            tm = JTemplate(template.file_name)
            file_name = tm.render(args)
        else:
            model_element_first_key = list(model.keys())[0]
            file_name = template.name + "_" + model[model_element_first_key]

        if template.type:
            file_name = file_name + "." + template.type

        return file_name

    def render_file_path(self, template: TemplateSchema, model) -> str:
        """Render file path

        Args:
            template (dict): The Template to generate.
            model (str): The model from which we want to generate.

        Returns:
            (str): the file path
        """
        filepath = ""
        args = {"model": SingleModelView(model), "full_model": self.model}
        if len(template.file_path) > 0:
            tm = JTemplate(template.file_path)
            filepath = tm.render(args)
        return filepath

    def render_model(
        self, template: TemplateSchema, model_element: Dict[str, str], from_list
    ):
        model_element_first_key = list(model_element.keys())[0]
        ret_value = "Nothing as been generated."
        if model_element_first_key:
            try:
                jinja_template = self.jinja_env.get_template(template.name)
                args = {
                    "model": model_element
                    if from_list
                    else SingleModelView(model_element),
                    "full_model": self.model,
                }
                ret_value = jinja_template.render(args)
            except TemplateNotFound:
                self.last_error = f"Could not find template in {template.name}"
                ret_value = self.last_error
                logging.error(self.last_error)
            except Exception as ex:
                ret_value = f"Exception in {template.name}: {str(ex)}"
        else:
            ret_value = "Part of the Model not found or empty"
        return {
            "name": self.render_file_name(template, model_element),
            "filepath": self.render_file_path(template, model_element),
            "content": ret_value,
        }

    def _render_template(self, name: str) -> List[Dict[str, str]]:
        output = []
        template = self.get_template_by_name(name)
        model_part = self.model
        path_steps = [elem for elem in template.path.split("/") if elem]

        for step in path_steps:
            try:
                model_part = self.model[step]
            except Exception:
                output.append(
                    {
                        "name": name,
                        "filepath": template.file_path,
                        "content": f"ERR:Given Path {template.path} ist not Valid:\
                               Step {step} not found.",
                    }
                )
                return output

        if isinstance(model_part, list):
            for element in model_part:
                output.append(self.render_model(template, element, from_list=True))
        elif isinstance(model_part, dict):
            output.append(self.render_model(template, model_part, from_list=False))
        else:
            output.append(
                {
                    "name": name,
                    "filepath": template.file_path,
                    "content": "ERR: "
                    + name
                    + " Model Path contains not an List or Dictionary.",
                }
            )
        return output

    def render_templates(self) -> List[Dict[str, str]]:
        rendered_outputs = []

        for template in self.configs.templates:
            output = []
            if not template.is_macro and not template.is_base:
                output = self._render_template(template.name)
                rendered_outputs.extend(output)

        return rendered_outputs

    def render(self) -> List[File]:
        """the render method renders a deployment group."""

        result = []
        for elem in self.render_templates():
            name = elem.get("name")
            content = elem.get("content")
            filepath: str = elem.get("filepath")
            if filepath == "/" or filepath == "./" or filepath == ".":
                root = self.output
            else:
                if filepath.startswith("/"):
                    root = self.output / Path(filepath[1:])
                else:
                    root = self.output / Path(filepath)
            file = File(name=name, root=root, content=content)
            result.append(file)

        return result

    def render_template(self, template_name):
        """The render method renders one template"""
        template = self.get_template_by_name(template_name)
        if template:
            return self._render_template(template.name)
        return {"name": f"ERR: {template_name}", "content": "template not found."}

    def _list_templates(self):
        """List the templates in the actual jinja environment"""

        return self.jinja_env.list_templates()



import argparse

from koodu.scripts.checks import check_templates, list_command
from koodu.scripts.generate import generate


def print_koodu():
    print("╔═══════════════════════════════════════════════════╗")
    print("║                  KOODU CLI TOOLS                  ║")
    print("║   Simple and lightweight code generator engine.   ║")
    print("╠═══════════════════════════════════════════════════╣")
    print("║ Options:                                          ║")
    print("║                                                   ║")
    print("║ 1. generate                                       ║")
    print("║ 2. list                                           ║")
    print("║ 3. check                                          ║")
    print("╚═══════════════════════════════════════════════════╝")
    print()


def main():
    parser = argparse.ArgumentParser(
        description="koodu CLI Tool - Simple and lightweight code generator engine.",
        usage="koodu [options] arguments",
    )
    subparser = parser.add_subparsers(dest="command", help="The command to execute")
    gen = subparser.add_parser("generate")
    check = subparser.add_parser("check")
    _list = subparser.add_parser("list")

    gen.add_argument(
        "-t",
        "--templates",
        type=str,
        required=True,
        help="The Path to the folder that contents the template's files.",
    )
    gen.add_argument(
        "-m",
        "--model",
        type=str,
        required=True,
        help="The Path to the model file(it should be a json file.)",
    )
    gen.add_argument(
        "-o", "--output", type=str, required=True, help="The Path to the output files."
    )

    check.add_argument(
        "--folder",
        "-f",
        type=str,
        required=True,
        help="The Path to the folder that contents the template's files.",
    )

    _list.add_argument(
        "option",
        type=str,
        choices=["models", "m", "templates", "t"],
        help="The option: what you want to list(model(m) or template(t))",
    )

    args = parser.parse_args()

    if args.command == "generate":
        generate(args)
    elif args.command == "check":
        output = check_templates(args)
        print("\n".join(output))
    elif args.command == "list":
        output = list_command(args)
        print("\n".join(output))


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import List

from jinja2 import Environment
from jinja2.exceptions import TemplateSyntaxError


def get_files_from_folder(folder: Path = Path(__file__)) -> List[Path]:
    """Recusively get all template in a Folder
    Args:
        folder (Path): The path to the folder.

    Return: The list of all template Path.
    """
    return [path for path in folder.rglob("**/*") if "jinja" in path]


def check_all_template(templates: List[Path]) -> List[bool]:
    """Check if all template in a list are valid jinja Template.

    Args:
        templates (List(Path)): the liste of template

    Returns:
        result (List(bool)): the list that content boolen representing
                             if each template in the list is valid or not.
    """
    result = []
    env = Environment()

    for path in templates:
        try:
            with open(path, "r") as fp:
                template = fp.read()

            env.parse(template)

            result.append(True)
        except TemplateSyntaxError:
            result.append(False)

    return result


def convert_symbols(symbols: List[bool]) -> List[str]:
    """Convert a list if bolean to a list of emoji.

    Args:
        symbols (List(bool)): the list of symbols

    Returns:
        emojis (List(str)): the list of corresponding emoji.
    """
    emojis = []
    for sym in symbols:
        if sym:
            emojis.append("\U00002705")
        else:
            emojis.append("\U0000274C")

    return emojis



from enum import Enum


class ModelTypeEmum(str, Enum):
    string = "String"
    integer = "Integer"
    bollean = "Boolean"
    date_and_time = "DateTime"
    just_date = "Date"
    ref = "reference"


class KooduTemplate(Enum):
    DJANGO = "django"
    FLASK = "flask"
    FASTAPI = "fastapi"
    EXPRESS = "express"
